/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * graph construction filter
 */

#ifndef CFILTERGRAPH_HH
#define CFILTERGRAPH_HH

#include <kdtree++/kdtree.hpp>
#if !KDTREE_VERSION || KDTREE_VERSION < 700
	#error "libkdtree++ >= 0.70 necessary"
#endif

#include "CVector.hh"
#include <iostream>
#include "CGraphRaw.hpp"
#include "CGraph.hpp"
#include "CFilterParticles.hpp"
#include "CPointRadius.hh"
#include <list>

/**
 * Convert particles to a graph
 *
 * There are different ways for data storage
 * Our implementation is using the kd tree.
 *
 * Because we work on a discretized grid, we can also store pointers to such a grid.
 * Of course this needs more memory could produce a performance improvement on small search areas.
 *
 * The advantage of the kdtree is that we can search very fast for neighbored pixels independend
 * of the grid resolution which may increase in the future. its also much faster if the algo
 * is extended to 3D
 *
 * A usual graph is constructed from 3d data within a few milliseconds
 */

#define MAX_STRIPS	20


template <typename T>
class CFilterGraph
{
	// gradient data handler to store for advanced particle searching
	CData<CVector<3,T> > *particles_gradient;

public:
	/**
	 * class to handle a point in the 3d space
	 */
	class CPoint
	{
	public:
		CPointRadius<3,T> point;
		bool accessed;	// this is set to true if the point was already accessed

		inline T &operator[](int i)
		{
			return point.pos[i];
		}

		inline CPoint()
		{
			point.pos.setZero();
			point.radius = 0;
		}

		inline CPoint(CPointRadius<3,T> &p)
		{
			point = p;
			accessed = false;
		}

		static bool cmp_sorted_points(CPoint *p1, CPoint *p2)
		{
			return (*p1).point.radius > (*p2).point.radius;
		}

	};

	CGraphRaw<3,T> strips;
	CGraph<3,T> graph;


	static bool save(CGraphRaw<3,T> &strips, CMemoryContainer &container)
	{
		return strips.save(container);
	}

	static bool load(CGraphRaw<3,T> &strips, CMemoryContainer &container)
	{
		return strips.load(container);
	}



private:
	CVector<3,int> max_area;		// area measurement
	CVector<3,int> max_border;		// max values minus ignore_start_border_size

	float min_next_distance;		// minimum distance to next node while constructing strips
	float min_next_distance_pow2;
	float max_next_distance;		// maximum distance to next node while constructing strips
	float max_next_distance_pow2;
	float connect_strip_distance_scale;

	float ignore_start_border_size;		// size of border where the strip cannot start

	/**
	 * list with points sorted by radius
	 * this is used to get the particle with the maximum radius to create the next strip
	 */
	list<CPoint*> radius_sorted_points;

	/*
	 * KD-TREE stuff
	 */
	inline static T tree_access_control(CPoint* p, int k)	{ return (*p)[k]; }

	typedef KDTree::KDTree<	3,		// dimension
				CPoint*,	// storage data type
				std::pointer_to_binary_function<CPoint*,int,float>	// access function to data in storage type { float (CPoint*)[int] }
			> kdtree_type;

	kdtree_type *tree;

	/**
	 * return true if point is within the border area or out of area
	 */
	bool inBorderArea(CVector<3,T> p)
	{
		return	p[0] < ignore_start_border_size	||
			p[1] < ignore_start_border_size	||
			p[2] < ignore_start_border_size	||
			p[0] > max_border[0]	||
			p[1] > max_border[1]	||
			p[2] > max_border[2]
			;
	};


public:
CFilterGraph()
{
	setParameters();
	tree = new kdtree_type(std::ptr_fun(tree_access_control));
}


/**
 * set parameters
 * \param p_min_next_distance	minimum distance to next node
 * \param p_max_next_distance	maximum distance to next node
 * \param p_ignore_start_border_size	ignore the nodes within this distance from the border by searching the strip start pixels
 * \param p_connect_strip_distance_scale	the "jump" distance is this value multiplied with the p_min_next_distance to connect one strip with another
 */
void setParameters(	float p_min_next_distance = 5.0f,
			float p_max_next_distance = 15.0f,
			float p_ignore_start_border_size = 3.0f,
			float p_connect_strip_distance_scale = 2.0f
		)
{
	min_next_distance = p_min_next_distance;
	min_next_distance_pow2 = min_next_distance*min_next_distance;

	max_next_distance = p_max_next_distance;
	max_next_distance_pow2 = max_next_distance*max_next_distance;

	ignore_start_border_size = p_ignore_start_border_size;
	connect_strip_distance_scale = p_connect_strip_distance_scale;
}

public:


void clear_tree()
{
	for (typename kdtree_type::iterator i = tree->begin(); i != tree->end(); i++)
		delete (*i);

	tree->clear();
}


void clear()
{
	clear_tree();	// clear the tree
	radius_sorted_points.clear();
	strips.clear();
}

/**
 * start with a new strip at start_point
 * \param start_point	the point where to start building the strip
 * \param strip		the strip to store the node data
 */
void filter_new_strip(CPoint &start_point, list< CNodeRaw<3,T> > &strip)
{
	/**
	 * if the strip is not empty, this is the 2nd call on the same strip
	 * to try to lengthen the strip to the opposite direction (see recursive call
	 * at the end of this function)
	 */
	bool empty_strip = strip.empty();

	start_point.accessed = true;
	if (empty_strip)
	{
		// insert the start point immediately at beginning of list
		strip.push_front(CNodeRaw<3,T>(start_point.point.pos, start_point.point.radius));	// strip makes a copy of start_point!
	}

	CNodeRaw<3,T> *prev_point = 0;
	float prev_point_dist_2 = 0.0f;	// squaredistance to previous point
	if (!empty_strip)
	{
		/**
		 * if the strip is not empty this is the try to lengthen the strip
		 * there are at least 2 nodes stored at the strip and we want to have
		 * the node before the last node (the one before start_node)
		 */
		prev_point = &(*(++(strip.begin())));
		prev_point_dist_2 = (prev_point->pos).dist(start_point.point.pos);
	}

	CPoint *search_point = &start_point;	// we start searching with the start_point
	vector<CPoint*> v;			// storage for the points within the search radius

	bool point_inserted = false;		// set to true if we insert another point than start_point
	bool border_touched = false;		// true if one pixel touched the border
	while(true)
	{
		/**
		 * first we start by searching for neighbored nodes within the radius |min_next_distance|
		 * an alternative approach would be to set the search radius to the radius of the sphering filter
		 *
		 * info - back_inserter: function to insert elements at the end of class
		 */
		tree->find_within_range(search_point, max_next_distance, std::back_inserter(v));

		/**
		 * the next step is to find the maximum node while removing the found nodes
		 */
		CPoint *max_node = 0;	// maximum node within the search radius
		float max_value = -1.0f;

		for (typename vector<CPoint*>::iterator i = v.begin(); i != v.end(); i++)
		{
			CPoint &test_point = *(*i);	// for better access

			if (&test_point == &start_point)	// avoid going back to the same point
				continue;

			if (prev_point)
			{
				/*
				 * if prev_point is set, the strip was not empty as this function was called.
				 * this function call tries to lenghten a previous strip and starts at the
				 * start_point of the previous call.
				 * the problem is that the surrounding points of start_point are already set
				 * as accessed. even if not, we are only interested in the points within the search 
				 * radius of start_point but not of those within the search radius of prev_point!!!
				 *
				 * thus we search in the point set out of the radius of prev_point
				 */
				CVector<3,T> delta = test_point.point.pos - prev_point->pos;

				if (delta.length2() < max_next_distance_pow2)
					continue;
			}
			else
			{
				if (test_point.accessed)
					continue;
			}

			// abort the stripping loop if one point touches the border
			if (point_inserted)
				if (inBorderArea(test_point.point.pos))
					border_touched = true;

			// calculate distance between this point and the local search_point
			CVector<3,T> delta = test_point.point.pos - search_point->point.pos;
			if (delta.length2() >= min_next_distance_pow2-0.5)	// subtract 0.5 because of discretized data
			{
				if (test_point.point.radius > max_value)
				{
					max_node = (*i);
					max_value = test_point.point.radius;
				}
			}

			test_point.accessed = true;
		}

		prev_point = 0; // set prev_point to 0 to disable prev_point handling after the first loop

		if (max_value < 0.0f)
			break;

		point_inserted = true;
		strip.push_front(CNodeRaw<3,T>(max_node->point.pos, max_node->point.radius));
		search_point = max_node;
		std::cout << "." << std::flush;

		if (border_touched)
			break;
	}

	if (!point_inserted)
	{
		/**
		 * this is the worst case which can happen: the strip is totally empty
		 * this happens if a node is totally alone at some place (hopefully with
		 * a lot of beer :-) )
		 */
		return;
	}

	if (empty_strip)
	{
		/**
		 * the strip was empty as this function was called
		 * so it can be that we can continue the strip on the other side because
		 * the point is searched by the maximum radius and the strip can continue
		 * to the "opposite direction" of the just generated strip
		 *
		 * so we go back to the first point and try to find a longer strip
		 */

		/* 1st step: reverse the strip to create correct connection */
		strip.reverse();

		/* 2nd: try to create the new strip */
		filter_new_strip(start_point, strip);
	}
}

void fixReverseIterators(list< CNodeRaw<3,T> > &node_list)
{
	/**
	 * iterate through all nodes in this list
	 */
	for (typename list< CNodeRaw<3,T> >::iterator i = node_list.begin(); i != node_list.end(); i++)
	{
		CNodeRaw<3,T> &n = (*i);
		if (n.neighbor != NULL)
		{
			CNeighborNode<3,T> &nn = *(n.neighbor->node->neighbor); 
			if (nn.node_list == &node_list)
			{
				/**
				 * if there is a neighbor we try to fix the reverse iterator
				 * bcz. it's damaged by the insert operation
				 */
				typename list< CNodeRaw<3,T> >::reverse_iterator i2;
				for (i2 = node_list.rbegin(); i2 != node_list.rend(); i2++)
				{
					if (nn.node == &*i2)
						{
						nn.node_rev_iter = i2;
						break;
					}
				}

				if (i2 == node_list.rend())
				{
					std::cerr << "FATAL ERROR 3: node not found" << std::endl;
					break;
				}
			}
		}
	}
}

/**
 * search for a new connection
 */
void search_connection(	CNodeRaw<3,T> &first_node,		// first node at an end of strip
			CNodeRaw<3,T> &prev_node,		// node next to the first node
			list< CNodeRaw<3,T> > &connect_strip,	// strip of this node to connect
			CFilterParticles<T> &filterParticles	// filter partifcles class for advances particle handling
		)
{
	if (first_node.neighbor)	// return if we already have a neighbor! todo: do something else
		return;

	// TODO: use more previous nodes (less weighted than prev_node) for direction
	CVector<3,T> delta = first_node.pos - prev_node.pos;
	float norm = 1.0f / delta.length();

	CVector<3,T> d;
	CPoint search_node;

	float scale = norm*max_next_distance*connect_strip_distance_scale;
	delta *= scale;
	d = CVector<3,T>(first_node.pos + delta);
	search_node.point.pos = d;

	if (!filterParticles.particle_movement(search_node.point.pos, *particles_gradient))
		return;

	if (inBorderArea(d))
		return;
	/*
	 * now we have to search for nodes within the distance max_next_distance away from the search_node
	 * because the tree data isnt linked to the data in the strips, we search within all strips for a possible match
	 */

	/*
	 * we use a search rectangle to fasten the search
	 */
	float x_min = search_node[0]-max_next_distance;
	float x_max = search_node[0]+max_next_distance;
	float y_min = search_node[1]-max_next_distance;
	float y_max = search_node[1]+max_next_distance;
	float z_min = search_node[2]-max_next_distance;
	float z_max = search_node[2]+max_next_distance;

	bool found = false;
	for (typename list< list< CNodeRaw<3,T> > >::iterator si = strips.begin(); si != strips.end(); si++)
	{
		if (&connect_strip == &(*si))	// dont connect to itself
			continue;

		list< CNodeRaw<3,T> > &node_list = *si;

		float min_dist_pow2 = std::numeric_limits<float>::max();
		CNodeRaw<3,T> *nearest_node = 0;
		typename list< CNodeRaw<3,T> >::iterator nearest_iter;

		for (typename list< CNodeRaw<3,T> >::iterator li = node_list.begin(); li != node_list.end(); li++)
		{
			CNodeRaw<3,T> &node = *li;
			if (	node[0] > x_min &&
				node[0] < x_max && 
				node[1] > y_min &&
				node[1] < y_max	&&
				node[2] > z_min &&
				node[2] < z_max	)
			{
				delta = node.pos - search_node.point.pos;
				float dist_pow2 = delta.length2();
				if (max_next_distance_pow2 > dist_pow2)
				{
					// calculate distance between found and first node
					delta = node.pos - first_node.pos;
					dist_pow2 = delta.length2();

					if (min_dist_pow2 > dist_pow2)
					{
						found = true;
						nearest_node = &node;
						nearest_iter = li;
						min_dist_pow2 = dist_pow2;
					}
				}
			}
		}

		if (found)
		{
			if (nearest_node->neighbor != NULL)
			{
				/*
				 * this node has already a neighbor
				 * => insert fake neighbor
				 */
				typename list< CNodeRaw<3,T> >::iterator move_iter = nearest_iter;
				nearest_iter = node_list.insert(nearest_iter, CNodeRaw<3,T>(*nearest_node));
				nearest_node = &*nearest_iter;

				fixReverseIterators(node_list);
			}


			// determine forward iterators
			typename list< CNodeRaw<3,T> >::iterator first_iter = connect_strip.begin();
			for (; first_iter != connect_strip.end(); first_iter++)
				if (&*first_iter == &first_node)
					break;

			// determine reverse iterators
			typename list< CNodeRaw<3,T> >::reverse_iterator first_rev_iter = connect_strip.rbegin();
			for (; first_rev_iter != connect_strip.rend(); first_rev_iter++)
				if (&*first_rev_iter == &first_node)
					break;
			if (first_rev_iter == connect_strip.rend())
			{
				std::cerr << "unknown error 1 (fatal)" << std::endl;
				return;
			}

			// determine reverse iterator for nearest iter
			typename list< CNodeRaw<3,T> >::reverse_iterator nearest_rev_iter = node_list.rbegin();
			for (; nearest_rev_iter != node_list.rend(); nearest_rev_iter++)
				if (&*nearest_rev_iter == nearest_node)
					break;
			if (nearest_rev_iter == node_list.rend())
			{
				std::cerr << "unknown error 2 (fatal)" << std::endl;
				return;
			}

			nearest_node->neighbor = new CNeighborNode<3,T>(connect_strip, first_iter, first_rev_iter, first_node);
			first_node.neighbor = new CNeighborNode<3,T>(node_list, nearest_iter, nearest_rev_iter, *nearest_node);

			/**
			 * if that node was the only one maybe we can go further in that direction
			 */
			if (node_list.size() == 1)
			{
				// insert fake node directing in the same direction
				CVector<3,T> nd = nearest_node->pos - first_node.pos;	// direction
				float scale = 1.0f / nd.length();
				nd *= scale;
				CNodeRaw<3,T> new_node(nearest_node->pos);
				new_node.pos += nd;

				node_list.push_back(new_node);
				fixReverseIterators(node_list);

				std::cout << "> " << node_list.back() << std::endl;
				search_connection(	
						node_list.back(),	// node next to the first node
						node_list.front(),	// node next to the first node
						node_list,		// strip of this node
						filterParticles		// particle filter for advanced particle handling
				);
			}

			break;
		}// if (found)
	}
	return;
}



/**
 * start the filter with point_list
 *
 *
 * TYPES:
 * the classes and parameters are a little bit confusing:
 *
 * vec3f	this vector is used as a list to get the input point ist
 * 		the first 2 components are the 2d coordinate ant the 3rd component the radius
 *
 * CPoint	the input list is stored in the class CPoint
 * 		CPoint specifies a point in the 2d space together with the radius and the information
 * 		if it was accessed previously
 *
 * CNodeRaw	this is the output type
 * 		the node stores the 2d position, the radius and maybe a connection node
 *
 * STORAGE:
 * the points are stored within the tree and in the sorted list as a pointer reference to save memory
 *
 * \param point_list	list with 2d points and the according sphere radius at the point position
 * \param p_max_area	maximum area of particles
 * \param filterParticles	handler to particleFilter with particles
 * \param p_particles_gradient	gradient for connection tracing
 */
void filter(	list< CPointRadius<3,T> > &point_list,
		CVector<3,int> p_max_area,
		CFilterParticles<T> &filterParticles,
		CData< CVector<3,T> > &p_particles_gradient
)
{
	particles_gradient = &p_particles_gradient;

	// setup
	max_area = p_max_area;
	max_border = max_area - ignore_start_border_size;

	clear();	// clear existing graphs

	// abort if the point list is empty (no input data)
	if (point_list.empty())
		return;

	std::cout << " kdtree(" << flush;

	/**
	 * insert point list into KD TREE and to point list sorted by radius
	 */
	for (typename list< CPointRadius<3,T> >::iterator iter = point_list.begin(); iter != point_list.end(); iter++)
	{
		CPointRadius<3,T> &p = *iter;

		CPoint *ip = new CPoint(p);	// TODO use radius
		tree->insert(ip);

		// dont start at border because those values may be incorrect!
		if (!inBorderArea(p.pos))
			radius_sorted_points.push_front(ip);
	}

	std::cout << "." << std::flush;
	radius_sorted_points.sort(CPoint::cmp_sorted_points);	// sort the points by radius, larger radii first

	std::cout << "." << std::flush;
	tree->optimize();		// optimize the KD TREE

	std::cout << ")" << std::endl << std::flush;

	/**
	 * now we always search for an untouched (un-accessed) point with the
	 * maximum radius and start the tracing
	 *
	 * IMPORTANT:
	 * we dont send SHORT STRIPS to trash because they can build up the connection to big spheres (because maybe loneley particles gather at the sphere center)!!!
	 */
	for (typename list<CPoint*>::iterator i = radius_sorted_points.begin(); i != radius_sorted_points.end(); i++)
	{
		CPoint &p = **i;

		if (p.accessed)
			continue;

		// create new strip list
		strips.push_back(list< CNodeRaw<3,T> >());

		// start strip at the point with the maximum radius
		filter_new_strip(p, *(strips.rbegin()));
	}

	radius_sorted_points.clear();	// not needed anymore

	/**
	 * Now we have a nice trackline between the particles but there one problem: the ending of the lines should be
	 * connected to the counterparts e.g. if there's a Y or X forking there's always a end of a line at such a fork
	 * because it's impossible that there are no endings at a X fork because the points laying at the center of this fork
	 * are inaccessible after the first strip was created (presumed that the particles lie within the search radius and
	 * thus are set to accessed)
	 *
	 * There's also a second problem: if the tube ends in a big hole, there are no particles left we can connect to
	 * (because the particles follow the gradient to the center of the big hole) but the trick is to go a few steps
	 * (e.g. the distance of the radius search) in the direction of the line ending and then trace the particle to the
	 * gradient center. if the particle arrived at the gradient center we can check if there are more particles within
	 * the radius and thus connect the line ending to the line in the center - IF there was created a strip!!!
	 * 
	 * HELL YEAH
	 */

/*
	std::cout << "PRE CHECK" << endl;
	checkNeighbors();
	std::cout << "PRE CHECK END" << endl;
*/
	// check every line ending for a connection
	for (typename list< list< CNodeRaw<3,T> > >::iterator i = strips.begin(); i != strips.end(); i++)
	{
		CNodeRaw<3,T> *n1, *n2;
		list< CNodeRaw<3,T> > &l = *i;

		// if there is just one node, we are not able to compute a direction to emit the virtual particle
		if (l.size() < 2)	continue;

		// load first and next node
		typename list< CNodeRaw<3,T> >::iterator in = l.begin();
		n1 = &(*in);
		n2 = &(*(++in));

		search_connection(*n1, *n2, l, filterParticles);	// start search

		// load last and previous node
		typename list< CNodeRaw<3,T> >::reverse_iterator rin = l.rbegin();
		n1 = &(*rin);
		n2 = &(*(++rin));

		search_connection(*n1, *n2, l, filterParticles);	// start search

		// try to find circular connections
		// the connection needs to have a minimum length of 4 times the search radius
		typename list< CNodeRaw<3,T> >::iterator prev_node = l.begin();
		if (prev_node != l.end())
		{
			typename list< CNodeRaw<3,T> >::iterator node = prev_node;
			node++;

			/**
			 * the edge needs to has a minimum length of max_next_distance_pow*4
			 */
			float dist = 0;
			while (node != l.end())
			{
				dist += (*node).pos.dist((*prev_node).pos);

				if (max_next_distance_pow2*16 < dist*dist)
					break;

				node++;
				prev_node++;
			}

			if (max_next_distance_pow2*16 < dist*dist)
			{
				// check if both endings
				CNodeRaw<3,T> *p_front = &l.front();
				CNodeRaw<3,T> *p_back = &l.back();

				float end_dist2 = (*p_front).pos.dist2((*p_back).pos);
				if (end_dist2 < max_next_distance_pow2)
				{
					// insert new node if neighbor is already used
					if ((*p_front).neighbor)
					{
						l.push_front(*p_front);
						p_front = &l.front();
					}

					if ((*p_back).neighbor)
					{
						l.push_back(*p_back);
						p_back = &l.back();
					}

					// front neighbor
					typename list< CNodeRaw<3,T> >::iterator front_iter = l.begin();
					typename list< CNodeRaw<3,T> >::reverse_iterator front_rev_iter = l.rbegin();
					while(&*front_iter != &*front_rev_iter)
						front_rev_iter++;

					// back neighbor
					typename list< CNodeRaw<3,T> >::iterator back_iter = l.begin();
					typename list< CNodeRaw<3,T> >::reverse_iterator back_rev_iter = l.rbegin();
					while(&*back_iter != &*back_rev_iter)
						back_iter++;

					(*p_front).neighbor = new CNeighborNode<3,T>(l, back_iter, back_rev_iter, *p_back);
					(*p_back).neighbor = new CNeighborNode<3,T>(l, front_iter, front_rev_iter, *p_front);
				}
			}
		}
	}

	// final check
//	std::cout << "FINAL CHECK" << endl;
	checkNeighbors();
//	std::cout << "FINAL CHECK END" << endl;

	/**
	 * now we remote small strips
	 */
	/*
	 * NOO!!! DONT REMOVE SMALL STRIPS! AT LEAST NOT IN 2D and 3D :) !!!
	 *
	for (list<list< CNodeRaw<N,T> >>::iterator i = strips.begin(); i != strips.end(); i++)
	{
		list< CNodeRaw<N,T> > &l = *i;

		if (l.size() <= 3)
		{
			bool noNeighbor = true;
			for (list< CNodeRaw<N,T> >::iterator ni = l.begin(); ni != l.end(); ni++)
			{
				if ((*ni).neighbor != NULL)
				{
					noNeighbor = false;
					break;
				}
			}

			if (noNeighbor)
			{
//				std::cout << "erase" << endl;
				i = strips.erase(i);
			}
		}
	}
	*/

	std::cout << " strips (" << strips.size() << ")" << endl;

	clear_tree();

	// convert strips represented by CRawGraph to CGraph
	strips.toGraph(graph);

	// output some fancy stuff to give the user the feeling that something was produced :-)

	int c = 0;
	int nodes = 0;
	for (typename list<typename CGraph<3,T>::CEdge>::iterator i = graph.edges.begin(); i != graph.edges.end(); i++)
	{
		typename CGraph<3,T>::CEdge &edge = *i;
		std::cout << c << ": "	<< edge.nodes.front().pos << "(" << edge.nodes.front().radius << ")\t"
		       			<< edge.nodes.back().pos << "(" << edge.nodes.back().radius << ")"
					<< "\t\tNeighs: " << edge.front_neighbors.size() << " " << edge.back_neighbors.size()
				       	<< "\t\tPoints on edge: " << edge.nodes.size() << endl;
		nodes += edge.nodes.size();
		c++;
	}
	std::cout << "|GraphPoints| = " << nodes << std::endl;

}


~CFilterGraph()
{
	clear_tree();
	delete tree;
}


/**
 * check if the neighbors have valid connections
 */
void checkNeighbors()
{
	for (typename list< list< CNodeRaw<3,T> > >::iterator i = strips.begin(); i != strips.end(); i++)
		for (typename list< CNodeRaw<3,T> >::iterator ni = (*i).begin(); ni != (*i).end(); ni++)
			if ((*ni).neighbor)
				(*ni).neighbor->checkNeighbor();

}

/**
 * Export filter for amira
 */
bool export_polyline(string &filename)
{
	std::ofstream file;
	file.open(filename.c_str());
	if (!file.is_open())
	{
		std::cerr << "failed to open file " << filename << std::endl;
		return false;
	}


	std::ostringstream vertices;
// data can be used to store radius
//	std::ostringstream data;
	std::ostringstream lines;

	int nLines = 0;
	int nVertices = 0;

	graph.renumEdges();
	for (typename list<typename CGraph<3,T>::CEdge>::iterator i = graph.edges.begin(); i != graph.edges.end(); i++)
	{
		typename CGraph<3,T>::CEdge &edge = *i;
		int start_vertex_id = nVertices;
		for (typename list<typename CGraph<3,T>::CNode>::iterator n = edge.nodes.begin(); n != edge.nodes.end(); n++)
		{
			typename CGraph<3,T>::CNode &node = *n;

			vertices << node.pos[0] << " " << node.pos[1] << " " << node.pos[2] << std::endl;
			lines << nVertices << " ";
			nLines++;	// increment nLines for each vertex!
			nVertices++;
		}
		int end_vertex_id = nVertices-1;

		lines << "-1" << std::endl;
		nLines++;

		// create line for front neighbors
		for (typename list<typename CGraph<3,T>::CEdgeNeighbor>::iterator nei = edge.front_neighbors.begin(); nei != edge.front_neighbors.end(); nei++)
		{
			typename CGraph<3,T>::CEdgeNeighbor &neighbor = *nei;
			if (neighbor.edge->nr > edge.nr)
				continue;

			lines << start_vertex_id << " ";
			nLines++;	// increment nLines for each vertex!

			if (!neighbor.reverse)
			{
				// connected to front neighbor of neighbored edge
				typename CGraph<3,T>::CNode &neigh_node = neighbor.edge->nodes.front();
				vertices << neigh_node.pos[0] << " " << neigh_node.pos[1] << " " << neigh_node.pos[2] << std::endl;
				lines << nVertices << " ";
				nVertices++;
				nLines++;	// increment nLines for each vertex!
			}
			else
			{
				// connected to front neighbor of neighbored edge
				typename CGraph<3,T>::CNode &neigh_node = neighbor.edge->nodes.back();
				vertices << neigh_node.pos[0] << " " << neigh_node.pos[1] << " " << neigh_node.pos[2] << std::endl;
				lines << nVertices << " ";
				nVertices++;
				nLines++;	// increment nLines for each vertex!
			}

			lines << "-1" << std::endl;
			nLines++;

		}

		// create lines for back neighbor
		for (typename list<typename CGraph<3,T>::CEdgeNeighbor>::iterator nei = edge.back_neighbors.begin(); nei != edge.back_neighbors.end(); nei++)
		{
			typename CGraph<3,T>::CEdgeNeighbor &neighbor = *nei;
			if (neighbor.edge->nr > edge.nr)
				continue;

			lines << end_vertex_id << " ";
			nLines++;	// increment nLines for each vertex!

			if (!neighbor.reverse)
			{
				// connected to front neighbor of neighbored edge
				typename CGraph<3,T>::CNode &neigh_node = neighbor.edge->nodes.front();
				vertices << neigh_node.pos[0] << " " << neigh_node.pos[1] << " " << neigh_node.pos[2] << std::endl;
				lines << nVertices << " ";
				nVertices++;
				nLines++;	// increment nLines for each vertex!
			}
			else
			{
				// connected to front neighbor of neighbored edge
				typename CGraph<3,T>::CNode &neigh_node = neighbor.edge->nodes.back();
				vertices << neigh_node.pos[0] << " " << neigh_node.pos[1] << " " << neigh_node.pos[2] << std::endl;
				lines << nVertices << " ";
				nVertices++;
				nLines++;	// increment nLines for each vertex!
			}

			lines << "-1" << std::endl;
			nLines++;

		}
	}

	file << "# AmiraMesh ASCII 1.0" << std::endl;
	file << std::endl;
	file << "define Lines " << nLines << std::endl;
	file << "define Vertices " << nVertices << std::endl;
	file << std::endl;
	file << "Parameters {" << std::endl;
	file << "    ContentType \"HxLineSet\"" << std::endl;
	file << "}" << std::endl;
	file << std::endl;
	file << "Vertices { float[3] Coordinates } = @1" << std::endl;
//	file << "Vertices { float Data } = @2" << std::endl;
	file << "Lines { int LineIdx } = @3" << std::endl;
	file << std::endl;
	file << "@1" << std::endl;
	file << vertices.str() << std::endl;
//	file << "@2" << std::endl;
//	file << std::endl;
//	file << data.str() << std::endl;
	file << "@3" << std::endl;
	file << lines.str() << std::endl;

	file.close();

	return true;
}

};

#endif
