/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __CFILTER_PARTICLES3_HPP
#define __CFILTER_PARTICLES3_HPP

/**
 * this filter emits particles which are at first aligned on a grid
 *
 * in each iteration step the particles follow the gradient 
 */

#include "CData.hpp"
#include "CVector.hh"
#include <list>
#include "CPointRadius.hh"
#include "CMemoryContainer.hpp"
#include <math.h>
#include <iostream>
#include <set>



template <typename T>
class CFilterParticles
{
public:
	class CParticle
	{
	public:
		CPointRadius<3,T> point;

		CParticle(T x, T y, T z, T r)
		{
			point.pos[0] = x;
			point.pos[1] = y;
			point.pos[2] = z;
			point.radius = r;
		}

		int steps;	// -1: follow gradient
				// 0: particle is stopped
				// >0: stepping is performed (steps is decreased in every iteration)
	};

public:
	list< CParticle > movingParticles;
	list< CPointRadius<3,T> > particleList;			// resoluting list with particle positions and sphere radius as 3rd component
	CVector<3,int> particle_domain;

	int init_distance;	// initial distance between particles
	T stop_gradient;	// value of gradient when stepping is activated
	T stop_gradient_pow_2;
	int start_steps;	// initial value of steps if stepping is activated
	int min_start_radius;		// mimimum radius

public:
	inline CFilterParticles()
	{
		reset();
	}

	inline void reset()
	{
		setParameters();
		movingParticles.clear();
		particleList.clear();
	}

	inline void setParameters(
				int p_init_distance = 3,		// initial distance between particles
				T p_stop_gradient = 2.5f,		// gradient when to stop the tracing (if leq than this value)
				int p_start_steps = 0,			// steps to perform particle tracing (usually unused - set to 0)
				int p_min_start_radius = 2		// minimum radius to emit particle at this position
			)
	{
		init_distance = p_init_distance;
		start_steps = p_start_steps;
		stop_gradient = p_stop_gradient;
		stop_gradient_pow_2 = p_stop_gradient*p_stop_gradient;
		min_start_radius = p_min_start_radius;
	}

	/**
	 * general filter
	 * gradientData and sphereData must have the same size!!!
	 */

	void filter(CData<CVector<3,T> > &p_gradientData, CData<T> &p_sphereData);

	/**
	 * \param p_gradientData	gradient data
	 * \param p_sphereData		used to setup the particle grid (according to min_start_radius)
	 */
	void filter_start(CData<CVector<3,T> > &p_gradientData, CData<T> &p_sphereData);

	/**
	 * compute one iteration
	 */
	bool filter_iteration(CData<CVector<3,T> > &src);

	/**
	 * compute the particle movements for all particles in one step
	 */
	void filter_iteration_fast(CData<CVector<3,T> > &src);

	/**
	 * call this if computations are finished
	 * the final particles (2 values) together with the sphere radius at the position
	 * are copied to the array particleList
	 * \param sphereData	this is the array with the sphere radius data
	 */
	void filter_finish(CData<T> &sphereData);


	/**
	 * get the movement of one particle within the gradient data field gradient_data
	 */
//	bool particle_movement(CParticle &particle, CData<CVector<3,T> > &gradient_data);
	bool particle_movement(CVector<3,T> &particle, CData<CVector<3,T> > &gradient_data);


	/**
	 * perform one iteration step on particle p
	 * \param p		2d coordinate of particle
	 * \param steps		maximum allowed steps for iteration - set to -1 to setup with start_steps
	 * \param gradient_data	the gradient data to trace the particle
	 */
//	inline bool iteration_step(CParticle &p, int *steps, CData<CVector<3,T> > &gradient_data);
	inline bool iteration_step(CVector<3,T> &p, int *steps, CData<CVector<3,T> > &gradient_data);


	static void save(list< CPointRadius<3,T> >& particleList, CMemoryContainer &container, CVector<3,int> &particle_domain)
	{
		int size = particleList.size();
		container.push(size);
		container.push(particle_domain[0]);
		container.push(particle_domain[1]);
		container.push(particle_domain[2]);

		for (typename list< CPointRadius<3,T> >::iterator i = particleList.begin(); i != particleList.end(); i++)
		{
			container.push((*i).pos[0]);
			container.push((*i).pos[1]);
			container.push((*i).pos[2]);
			container.push((*i).radius);
		}
	}


	static bool load(list< CPointRadius<3,T> >& particleList, CMemoryContainer &container, CVector<3,int> &particle_domain)
	{
		int size;	container.get(&size);
		container.get(&particle_domain[0]);
		container.get(&particle_domain[1]);
		container.get(&particle_domain[2]);

		particleList.clear();
		for (int i = 0; i < size; i++)
		{
			CPointRadius<3,T> v;
			container.get(&v.pos[0]);
			container.get(&v.pos[1]);
			container.get(&v.pos[2]);
			container.get(&v.radius);
			particleList.push_back(v);
		}
		return true;
	}
};



template < typename T>
void CFilterParticles<T>::filter_start(CData<CVector<3,T> > &p_gradientData, CData<T> &p_sphereData)
{
	particle_domain[0] = p_gradientData.w;
	particle_domain[1] = p_gradientData.h;
	particle_domain[2] = p_gradientData.d;

	movingParticles.clear();
	CVector<3,T> vec_null = CVector<3,T>(0.0);

	for (int z = 1; z < p_gradientData.d-1; z += init_distance)
		for (int y = 1; y < p_gradientData.h-1; y += init_distance)
			for (int x = 1; x < p_sphereData.w-1; x += init_distance)
				if (p_gradientData.get(x, y, z) != vec_null && p_sphereData.get(x, y, z) >= min_start_radius)
				{
					CParticle p(x,y,z,0);
					p.steps = -1;

					movingParticles.insert(movingParticles.begin(), p);
				}
}


template <typename T>
//inline bool CFilterParticles<T>::iteration_step(CParticle &p, int *steps, CData<CVector<3,T> > &gradient_data)
inline bool CFilterParticles<T>::iteration_step(CVector<3,T> &p, int *steps, CData<CVector<3,T> > &gradient_data)
{
	CVector<3,T> grad = gradient_data.get((int)p[0], (int)p[1], (int)p[2]);

	// normalize gradient movement (only 1 or max of 2 cells movement)
	/*
	T ax = fabsf(grad[0]);
	T ay = fabsf(grad[1]);
	T az = fabsf(grad[2]);

	T max = (ax > ay ? ax : ay);
	max = (max > az ? max : az);
	T scale = 1.0f / (max - 0.001);
	*/

	T length = grad.length();

	if (length < 0.5)
	{
		steps = 0;
		return true;
	}

	T scale = (T)1 / (length - (T)0.001);

	grad *= scale;

	if (length <= stop_gradient)	// change to stepping if gradient change (change in radius) remains almost constant (<= stop_gradient)
	{
		if (*steps == -1)
			*steps = start_steps;
	}

	/*
	// now discretize it and check that it's < 1
	int dx = (int)grad[0];
	int dy = (int)grad[1];
	int dz = (int)grad[2];

	if (dx > 1)		dx = 1;
	else if (dx < -1)	dx = -1;

	if (dy > 1)		dy = 1;
	else if (dy < -1)	dy = -1;

	if (dz > 1)		dz = 1;
	else if (dz < -1)	dz = -1;

	p[0] += dx;
	p[1] += dy;
	p[2] += dz;
	*/
	p += grad;

	if (!gradient_data.inCheck((int)p[0], (int)p[1], (int)p[2]))
		return false;

	if (*steps > 0)
		*steps--;

	return true;
}

template <typename T>
void CFilterParticles<T>::filter_iteration_fast(CData<CVector<3,T> > &src)
{
	for (typename list< CParticle >::iterator iter = movingParticles.begin(); iter != movingParticles.end(); iter++)
		particle_movement((*iter).point.pos, src);
}


/**
 * move a particle through the gradient data
 * \return	false if the particle is not usable (out of border)
 */
template <typename T>
//bool CFilterParticles<T>::particle_movement(CParticle &p, CData<CVector<3,T> > &gradient_data)
bool CFilterParticles<T>::particle_movement(CVector<3,T> &p, CData<CVector<3,T> > &gradient_data)
{
	if (!gradient_data.inCheck((int)p[0], (int)p[1], (int)p[2]))
		return false;

	int steps = -1;
	for (int iterations = (gradient_data.w+gradient_data.h)/2; iterations > 0; iterations--)
	{
		if (!iteration_step(p, &steps, gradient_data))
			return false;

		if (steps == 0)
			return true;
	}
	return true;
}



template <typename T>
bool CFilterParticles<T>::filter_iteration(CData<CVector<3,T> > &src)
{
	typename list< CParticle >::iterator iter = movingParticles.begin();

	bool movement = false;
	while(iter != movingParticles.end())
	{
		CParticle &p = *iter;

		if (p.steps == 0)	// particle is stopped
		{
			iter++;
			continue;
		}

		if (!iteration_step(p.pos, &(p.steps), src))
		{
			iter = movingParticles.erase(iter);
			iter++;
			continue;
		}

		movement = true;
		iter++;
	}
	return movement;
}



template <typename T>
void CFilterParticles<T>::filter(CData<CVector<3,T> > &p_gradientData, CData<T> &p_sphereData)
{
	filter_start(p_gradientData, p_sphereData);
	filter_iteration_fast(p_gradientData);
	filter_finish(p_sphereData);
}


/**
 * finish the filter:
 * 1st: store the data to set to avoid duplicate particles
 * 2nd: store the data back to a list
 */
template <typename T>
void CFilterParticles<T>::filter_finish(CData<T> &sphereData)
{
	typedef set<CPointRadius<3,T>, typename CPointRadius<3,T>::compareSet > set_nodes_type;
	set_nodes_type set_nodes;

	std::cout << " remove dups(" << movingParticles.size() << std::flush;
	/**
	 * the first step is to copy the particles to a set
	 * this avoids duplicate particles which share the same position,
	 * because the position is the compare criteria
	 */
	for (typename list< CParticle >::iterator i = movingParticles.begin(); i != movingParticles.end(); i++)
	{
		int x = (int)roundf((*i).point.pos[0]);
		int y = (int)roundf((*i).point.pos[1]);
		int z = (int)roundf((*i).point.pos[2]);

		if (	x == 0 || x == particle_domain[0]-1	||
			y == 0 || y == particle_domain[1]-1	||
			z == 0 || z == particle_domain[2]-1	)
			continue;

		if (sphereData.inCheck(x, y, z))
		{
			CVector<3,T> pos((T)x, (T)y, (T)z);
			CPointRadius<3,T> v(pos, sphereData.get(x, y, z));
			set_nodes.insert(v);	/* ignore if insert fails! */
		}
	}
	std::cout << " -> " << set_nodes.size() << ") " << std::flush;

	particleList.clear();
	for (typename set_nodes_type::iterator i = set_nodes.begin(); i != set_nodes.end(); i++)
		particleList.insert(particleList.begin(), (*i));

	movingParticles.clear();
}
#endif
