/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __CFILTERSPHERE_HPP
#define __CFILTERSPHERE_HPP

#include "CMemoryContainer.hpp"
#include <sys/time.h>

// use the fft as sphere filter

#define __CFILTER_SPHERE_USE_FFT	1

#if __CFILTER_SPHERE_USE_FFT
	#include <complex.h>
	#include <fftw3.h>
#endif

/**
 * 3 dimensional sphere filter
 */
template< typename T >
class CFilterSphere
{
	T min_value;
	T max_value;
	int min_radius;
	T valid_error;
	T one_minus_valid_error;
	int fft_threads;
	int max_radius;

	CVector<3,int> **sphere_points;	// sphere point positions
	int *sphere_points_count;	// count of sphere_points

public:

	CFilterSphere()
	{
		sphere_points = 0;
		setParameters();
		max_radius = -1;
	}

	CFilterSphere(int p_max_radius)
	{
		sphere_points = 0;
		setParameters();
		max_radius = p_max_radius-1;	// if both variables differ, the cache is created
		create_sphere_cache(p_max_radius);
	}

	~CFilterSphere()
	{
		if (sphere_points)
		{
			for (int i = 0; i < max_radius; i++)
				delete sphere_points[i];

			delete sphere_points;
			delete sphere_points_count;
		}
	}

	/**
	 * \param min_value	minimum value for windowing
	 * \param max_value	maximum value for windowing
	 * \param min_radius	minimum radius to allow the aborting
	 * \param valid_error	maximum error rate (pixels out of window) \in [0;1]
	 */
	inline void setParameters(	T p_min_value = -50,
					T p_max_value = 50,
					int p_min_radius = 3,
					T p_valid_error = (T)0.05,
					int p_fft_threads = 1
				)
	{
		min_value = p_min_value;
		max_value = p_max_value;
		min_radius = p_min_radius;
		valid_error = p_valid_error;
		one_minus_valid_error = (T)1-valid_error;
		fft_threads = p_fft_threads;
	}


	/*
	 * \param src	source data
	 * \param dst	destination data
	 */

	static bool save(CMemoryContainer &container, CData<T> &data)
	{
		container.push((int)data.w);
		container.push((int)data.h);
		container.push((int)data.d);
		for (size_t i = 0; i < data.pixel_size; i++)
			container.push(data.data[i]);
		return true;
	}

	static bool load(CMemoryContainer &container, CData<T> &data)
	{
		int w, h, d;
		container.get(&w);
		container.get(&h);
		container.get(&d);

		data.resize(w, h, d);

		for (size_t i = 0; i < data.pixel_size; i++)
			container.get(&(data.data[i]));
		return true;
	}


	// fast sphere filter using precached circumference data


	// initialize int field (better cache coherancy)

	/**
	 * preview filter for window
	 */
	void filter_preview(CData<T> &src, CData<T> &dst)
	{
		dst.resize(src.w, src.h, src.d);
		for (int z = 0; z < src.d; z++)
		{
			for (int y = 0; y < src.h; y++)
			{
				for (int x = 0; x < src.w; x++)
				{
					T value = src.get(x, y, z);
					dst.set(x, y, z, (value >= min_value && value <= max_value));
				}
			}
		}
	}

#if __CFILTER_SPHERE_USE_FFT
	/**
	 * analyse the output of the dct filter and set the radius in dst spherical filter output if appropriate
	 * \param out:	output of convolution in spatial room
	 * \param dst:	spherical filter output
	 */
	void analyse_output(float *out, CData<float> &dst, int radius, size_t *output_written_points_count)
	{
		// postprocess (correct scaling)
		// because the data is even, also divide by 2*2*2
		T inv_value = (T)1/(dst.pixel_size*8);

		for (size_t i = 0; i < dst.pixel_size; i++)
		{
			if (dst.data[i] < 0)
			{
				// output data was not set

				// matches_percent represents the precentage of the pixels
				// which were within the window values and within the radius surface
				T matched_percent = out[i]*inv_value;

				if (matched_percent < one_minus_valid_error)
				{
					// abort here and store radius
					dst.data[i] = (T)radius;
					(*output_written_points_count)++;
				}
			}
		}

		std::cout << "." << std::flush;
	}
#endif
	/**
	 * FAST FILTER
	 * this filter uses the precached radius data from points[radius]
	 *
	 * \param time_after_creating_plans	if this pointer is not NULL, store the system time to this value to get measurements after creating points (bad fuck-sucking hack)
	 */
	void filter(CData<float> &src, CData<float> &dst, double *time_after_creating_plans = NULL)
	{
		std::cout << src.getMin() << " " << src.getMax() << std::endl;

		// resize destination data
		dst.resize(src.w, src.h, src.d);

/**
 * use FFT FOR SPHERE FILTER
 */
#if __CFILTER_SPHERE_USE_FFT

		float *src_ft	= (float*)fftwf_malloc(sizeof(float)*src.pixel_size);
		float *kernel	= (float*)fftwf_malloc(sizeof(float)*src.pixel_size);
		float *kernel_ft	= (float*)fftwf_malloc(sizeof(float)*src.pixel_size);
		float *out_ft	= (float*)fftwf_malloc(sizeof(float)*src.pixel_size);
		float *out	= (float*)fftwf_malloc(sizeof(float)*src.pixel_size);

		/**
		 * output description for debug information
		 */
		std::cout << "---------------------" << std::endl;
		if (fft_threads > 1)
		{
			std::cout << "using " << fft_threads << " threads" << std::endl;
		}
		else
		{
			std::cout << "using no threading" << std::endl;
		}

		std::cout << "---------------------" << std::endl;
		std::cout << "m: mesaure" << std::endl;
		std::cout << "f: fourier" << std::endl;
		std::cout << "i: inverse fourier" << std::endl;
		std::cout << "k: preparing kernel" << std::endl;
		std::cout << "c: convolution" << std::endl;
		std::cout << "a: analyse output" << std::endl;
		std::cout << "---------------------" << std::endl;

		/**
		 * INIT THREADS
		 */
		if (fft_threads > 1)
		{
			fftwf_init_threads();			// use threads
			fftwf_plan_with_nthreads(fft_threads);	// initialize threads
		}

		std::cout << "[" << std::flush;

		/**
		 * MEASURE
		 * use MEASURE flag to determine the fastest method
		 */

		// create fftw plans
		std::cout << "  plans:" << std::flush;
		// estimate because we need it only once
		fftwf_plan p_src_fft = fftwf_plan_r2r_3d(src.d, src.h, src.w, dst.data, src_ft, FFTW_REDFT10, FFTW_REDFT10, FFTW_REDFT10, FFTW_MEASURE);
		std::cout << "." << std::flush;
		fftwf_plan p_kernel_fft = fftwf_plan_r2r_3d(src.d, src.h, src.w, kernel, kernel_ft, FFTW_REDFT10, FFTW_REDFT10, FFTW_REDFT10, FFTW_MEASURE);
		std::cout << "." << std::flush;
		fftwf_plan p_out_ifft = fftwf_plan_r2r_3d(src.d, src.h, src.w, out_ft, out, FFTW_REDFT01, FFTW_REDFT01, FFTW_REDFT01, FFTW_MEASURE);
		std::cout << "." << std::flush;


		if (time_after_creating_plans != NULL)
		{
			std::cout << "[updating extern timestamp]" << std::flush;
			timeval now;
			gettimeofday(&now, NULL);
			*time_after_creating_plans = (double)now.tv_usec/(double)1000000.0;
			*time_after_creating_plans += (double)now.tv_sec;
		}

		/**
		 * FLAG FIELD
		 * use dst data field to create T flag field for in/out of window
		 */
		std::cout << "  flag_field:" << std::flush;
		for (size_t i = 0; i < src.pixel_size; i++)
		{
			T value = src.data[i];
			dst.data[i] = (value >= min_value && value <= max_value);
		}
		std::cout << "." << std::flush;


		/**
		 * FLAG to FREQUENCY FLAG
		 */
		std::cout << "f" << std::flush;
		fftwf_execute(p_src_fft);
		std::cout << "." << std::flush;


		// set output radius to -1 (radius not yet set)
		for (size_t i = 0; i < dst.pixel_size; i++)
			dst.data[i] = (T)-1;


		/**
		 * KERNEL
		 */
		std::cout << "   0-" << min_radius << ":" << std::flush;

		std::cout << "k" << std::flush;
		for (size_t i = 0; i < dst.pixel_size; i++)
			kernel[i] = 0;

		int sphere_new_pixels = 0;

		for (int r = 0; r <= min_radius && r < max_radius; r++)
			sphere_new_pixels += sphere_points_count[r];

		sphere_new_pixels *= (T)8;

		T inv_sphere_new_pixels = (T)1/(T)sphere_new_pixels;

		// the first kernel we create is created by all the points up to min_radius
		for (int r = 0; r <= min_radius && r < max_radius; r++)
		{
			for (int i = 0; i < sphere_points_count[r]; i++)
			{
				CVector<3,int> &k_point = sphere_points[r][i];
				if (!src.inCheck(k_point))
					continue;

				kernel[k_point[0] + k_point[1]*src.w + k_point[2]*src.w*src.h] = inv_sphere_new_pixels;
			}
		}
		std::cout << "." << std::flush;

		/**
		 * DCT TRANSFORM KERNEL
		 */
		std::cout << "f" << std::flush;
		fftwf_execute(p_kernel_fft);
		std::cout << "." << std::flush;

		// clear points of start sphere to avoid clearing the whole dataset
		for (int r = 0; r <= min_radius && r <= max_radius; r++)
		{
			for (int i = 0; i < sphere_points_count[r]; i++)
			{
				CVector<3,int> &k_point = sphere_points[r][i];
				if (!src.inCheck(k_point))
					continue;
				kernel[k_point[0] + k_point[1]*src.w + k_point[2]*src.w*src.h] = 0;
			}
		}

		/**
		 * CONVOLUTION
		 */
		std::cout << "c" << std::flush;
		for (size_t i = 0; i < src.pixel_size; i++)
			out_ft[i] = kernel_ft[i]*src_ft[i];
		std::cout << "." << std::flush;

		/**
		 * FILTERED IMAGE TO SPATIAL DOMAIN
		 */
		std::cout << "i" << std::flush;
		fftwf_execute(p_out_ifft);
		std::cout << "." << std::flush;

		size_t output_written_points_count = 0;
		/**
		 * UPDATE OUTPUT DATA
		 */
		analyse_output(out, dst, 0, &output_written_points_count);	// set to 0 => radius = 0

		for (int r = min_radius+1; r < max_radius; r++)
		//for (int r = min_radius+1; r < 10; r++)
		{
			std::cout << "   " << r << ":" << std::flush;

			// start computing the kernel for sphere contents with min_radius
			std::cout << "k" << std::flush;

			T inv_sphere_new_pixels = (T)1/(T)(sphere_points_count[r]*8);

			for (int i = 0; i < sphere_points_count[r]; i++)
			{
				CVector<3,int> &k_point = sphere_points[r][i];
				if (!src.inCheck(k_point))
					continue;
				kernel[k_point[0] + k_point[1]*src.w + k_point[2]*src.w*src.h] = inv_sphere_new_pixels;
			}

			std::cout << "." << std::flush;

			/**
			 * DCT TRANSFORM KERNEL
			 */
			std::cout << "f" << std::flush;
			fftwf_execute(p_kernel_fft);
			std::cout << "." << std::flush;

			// clear previously flagged sphere surface points
			// this is (or should be) faster than erasing the whole domain
			for (int i = 0; i < sphere_points_count[r]; i++)
			{
				CVector<3,int> &k_point = sphere_points[r][i];
				if (!src.inCheck(k_point))
					continue;
				kernel[k_point[0] + k_point[1]*src.w + k_point[2]*src.w*src.h] = 0;
			}

			/**
			 * CONVOLUTION
			 */
			std::cout << "c" << std::flush;
			for (size_t i = 0; i < src.pixel_size; i++)
				out_ft[i] = kernel_ft[i]*src_ft[i];
			std::cout << "." << std::flush;

			/**
			 * FILTERED IMAGE TO SPATIAL DOMAIN
			 */
			std::cout << "i" << std::flush;
			fftwf_execute(p_out_ifft);
			std::cout << "." << std::flush;

			/**
			 * UPDATE OUTPUT DATA
			 */
			analyse_output(out, dst, r, &output_written_points_count);

			if (output_written_points_count == src.pixel_size)
			{
				std::cout << " early break" << std::endl;
				break;
			}
		}

		fftwf_destroy_plan(p_out_ifft);
		fftwf_destroy_plan(p_kernel_fft);
		fftwf_destroy_plan(p_src_fft);

		if (fft_threads > 1)
			fftwf_cleanup_threads();

		fftwf_free(src_ft);
		fftwf_free(kernel);
		fftwf_free(kernel_ft);
		fftwf_free(out_ft);
		fftwf_free(out);

#else
		// initialize flag field (better cache coherancy)

		CData<bool> flag_field(src.w, src.h, src.d);
		for (int z = 0; z < src.d; z++)
		{
			for (int y = 0; y < src.h; y++)
			{
				for (int x = 0; x < src.w; x++)
				{
					T value = src.get(x, y, z);
					flag_field.set(x, y, z, (value < min_value || value > max_value));
				}
			}
		}

		dst.resize(src.w, src.h, src.d);

		std::cout << "[" << std::flush;

		int percent = 0;
		for (int z = 0; z < src.d; z++)
		{
			for (int y = 0; y < src.h; y++)
			{
				for (int x = 0; x < src.w; x++)
				{
					CVector<3,int> middle_point(x,y,z);

					int sphere_mismatch = 0;	// mismatched pixels
					int sum = 0;		// sum of all checked pixels

					int radius = 0;
					for (; radius < max_radius; radius++)
					{
						int radius_mismatch = 0;	// mismatches for this radius
						for (int i = 0; i < sphere_points_count[radius]; i++)
						{
							CVector<3,int> point = middle_point + sphere_points[radius][i];
							if (!flag_field.inCheck(point))
								continue;

							if (flag_field.get(point))
								radius_mismatch++;
							sum++;
						}

						if (min_radius < radius)
						{
							/**
							 * the collected points up to min_radius are compared with the valid error
							 */
							if ((T)radius_mismatch / (T)sphere_points_count[radius] > valid_error)
								break;
						}
						else
						{
							sphere_mismatch += radius_mismatch;

							if (min_radius == radius)
							{
								/**
								 * for all the new points within the radius, we compare the error rate
								 */
								if ((T)sphere_mismatch / (T)(sum) > valid_error)
									break;
							}
						}
					}

					dst.set(x, y, z, (min_radius < radius ? radius : 0));
				}
			}

			if (100*(z-1)/(src.d-2) > percent)
			{
				std::cout << "." << std::flush;
				percent++;
			}
		}

		std::cout << "]" << std::flush;
#endif
	}


	void create_sphere_cache(int p_max_radius)
	{
		if (max_radius == p_max_radius)
			return;

		if (sphere_points)
		{
			for (int i = 0; i < max_radius; i++)
				delete sphere_points[i];

			delete sphere_points;
			delete sphere_points_count;
		}

		max_radius = p_max_radius;

		if (max_radius == 0)
			return;

		std::cout << "creating sphere cache for filter_sphere [" << std::flush;

		// TODO: better way would be to iterate over the points and insert the point at the appropriate radius position
		sphere_points = new CVector<3,int>*[max_radius];
		sphere_points_count = new int[max_radius];

		for (int radius = 0; radius < max_radius; radius++)
		{
			list< CVector<3,int> > radius_list;

			int radius_2_low_bound = radius*radius;
			int radius_2_high_bound = (radius+1)*(radius+1);

/*
 * __CFILTER_SPHERE_USE_FFT:
 * if we use the DCT, we dont need to store the whole sphere
 * because of the mirrored data on the borders using DCT, it's enough to store
 * a quarter of the sphere
 */

			// search within square around radius
#if __CFILTER_SPHERE_USE_FFT
			for (int z = 0; z <= radius; z++)
#else
			for (int z = -radius; z <= radius; z++)
#endif
			{
				// search within square around radius
#if __CFILTER_SPHERE_USE_FFT
				for (int y = 0; y <= radius; y++)
#else
				for (int y = -radius; y <= radius; y++)
#endif
				{
					bool entered_1st = false;	// set to true if radius is entered
#if __CFILTER_SPHERE_USE_FFT
					for (int x = 0; x <= radius; x++)
#else
					bool entered_2nd = false;	// set to true if radius is entered 2nd time
					for (int x = -radius; x <= radius; x++)
#endif
					{
						// only search on the circle border
						int radius_2 = z*z + y*y + x*x;

#if __CFILTER_SPHERE_USE_FFT
						if (radius_2 >= radius_2_high_bound)
						{
							// break using fft because we need just the quarter of the sphere
							break;
						}
#endif

						if (radius_2 >= radius_2_low_bound && radius_2 < radius_2_high_bound)
						{
		//					std::cout << "IN: " << x << " " << y << "       " << radius_2 << std::endl;
							entered_1st = true;

							// insert into list
							radius_list.push_back(CVector<3,int>(x, y, z));
						}
#if !__CFILTER_SPHERE_USE_FFT
						else
						{
							if (entered_1st && x < 0)	// we exit the left side of the circle, thus we jump directly to the right side
							{
								x = -x;
								entered_2nd = true;	// after this trick, we immediatly enter the right part of the circle
								continue;
							}

							if (!entered_2nd && x > 0)
								break;
						}
#endif
					}
				}
			}

			CVector<3,int>* sphere_list = new CVector<3,int>[radius_list.size()];
			sphere_points[radius] = sphere_list;
			sphere_points_count[radius] = radius_list.size();

			int pos = 0;
			for (list< CVector<3,int> >::iterator i = radius_list.begin(); i != radius_list.end(); i++)
			{
				sphere_list[pos] = *i;
				pos++;
			}

			std::cout << "." << std::flush;
		}

		std::cout << "] OK" << std::endl;
	}
};

#endif
